/*
Auf Basis eines RP2040-Pico-Boards wird ein 32 KB EPROM simuliert, indem die Adressleitungen
A0 - A14 (GPIO0 - GPIO14) permanent gelesen und die dazugehoerigen Daten 
aus einem rom_puffer[] an den Ausgaengen D0 - D7 (GPIO15 - GPIO22) ausgegeben werden.
Bei der gewaehlten Taktfrequenz des RP2040 von 200 MHz ergibt sich eine eff. ZUgriffszeit von 
<= 120 ns. Mit (problemloser) Uebertaktung auf 360 MHz lassen sich <= 70 ns erreichen.

Um Daten im FLASH-Speicher ablegen zu können und Kollisionen mit der Programmausfuehrung
zu vermeiden, wird das komplette Programm ins RAM nach RAMSTART kopiert. 
Dies ist der XIP-Cache, der nicht mehr benötigt wird.

Abschliessend muss die ISR-Vektortabelle angepasst werden und das Programm wird
erneut gestartet, wobei die Startadresse von main() im RAM angesprungen wird.

2025-04-21
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#include "rom_simul.h"

extern void rom_simulation(uint32_t maske);
extern uint8_t rom_puffer[];

extern void trigger_reset(void);
extern void init_PIO0_RX(void);
extern void init_PIO1(void);

uint32_t SystemCoreClock  = (DEF_SYSCLOCK);       // siehe simul.h
uint32_t F_xtal = MHZ_12;                         // onboard default


#define CODE_OFFSET (RAMSTART - CODESTART)
#define VEC_TAB_LEN (RTC_IRQ_IRQn - Reset_IRQn +1)

int main(void);

// Code in XIP-Cache umkopieren und Vektoren anpassen
void rom2ram(void)
{
volatile uint32_t *p, *q;
uint32_t i;
// Aufrufadresse von main() muss veraendert werden
int (*main_im_ram)(void) = (int(*)(void)) ((uint32_t) main + CODE_OFFSET);
static int ins_ram_kopiert;                       // einmalig nach Programmstart
  if(!ins_ram_kopiert) {
    XIP_CTRL->CTRL &= ~XIP_CTRL_CTRL_EN_Msk;      // cache abschalten
  
    p = (uint32_t *) (RAMSTART);
    q = (uint32_t *) (CODESTART);
    for(i = 0; i < CODESIZE_8KB/4; i++)           // 8 kB Programm umkopieren
      *p++ = *q++;
  
    p = (uint32_t *) (RAMSTART+0x100);            // neue Vektortabelle ab hier
    for(i = 0; i <= VEC_TAB_LEN; i++) {           // Vektoren anpassen
      if(*p >= CODESTART && *p <= CODESTART+CODESIZE_8KB)  
        *p += CODE_OFFSET;
      p++;  
    }  
    SCB->VTOR += CODE_OFFSET;                     // nicht vergessen!
    ins_ram_kopiert = 1;                          // erledigt
    main_im_ram();                                // Aufruf von main() mit neuer Adresse
  }                                     
}

void short_delay(void)                            // no timer, no ISR
{
  volatile int i = 1000;
  while(i--);
}  

// setup XOSC, SystemCoreClock and some IO-clock
void init_clock(void)
{
  VREG_AND_CHIP_RESET_SET-> VREG = 15<<VREG_AND_CHIP_RESET_VREG_VSEL_Pos;
  while(!(VREG_AND_CHIP_RESET->VREG & VREG_AND_CHIP_RESET_VREG_ROK_Msk));
  short_delay();                                              // waiting for stable Vreg 

  XOSC->CTRL     = (XOSC_CTRL_FREQ_RANGE_1_15MHZ << XOSC_CTRL_FREQ_RANGE_Pos);
  XOSC->STARTUP  = 100;                                       // waiting for stable XO 12 MHz
  XOSC_SET->CTRL = (XOSC_CTRL_ENABLE_ENABLE << XOSC_CTRL_ENABLE_Pos);
  while (!(XOSC->STATUS & XOSC_STATUS_STABLE_Msk));
  
  RESETS_CLR->RESET = RESETS_RESET_pll_sys_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pll_sys_Msk));


  PLL_SYS->CS = (REF_DIV << PLL_SYS_CS_REFDIV_Pos);           // PLL: 6 MHz / step
  PLL_SYS->FBDIV_INT = (SystemCoreClock*PLL_DIV2*REF_DIV)/(F_xtal/PLL_DIV1);  // PLL-VCO typ. 1310 MHz
 
  PLL_SYS->PRIM = (PLL_DIV1 << PLL_SYS_PRIM_POSTDIV1_Pos) |   // setup devider
                  (PLL_DIV2 << PLL_SYS_PRIM_POSTDIV2_Pos);
  SystemCoreClock = PLL_SYS->FBDIV_INT *(F_xtal/REF_DIV) /    // calculate eff. frequency
                    (PLL_DIV1*PLL_DIV2);

  PLL_SYS_CLR->PWR = PLL_SYS_PWR_VCOPD_Msk | PLL_SYS_PWR_PD_Msk;
  while (!(PLL_SYS->CS & PLL_SYS_CS_LOCK_Msk));               // wait until done

  PLL_SYS_CLR->PWR = PLL_SYS_PWR_POSTDIVPD_Msk;

  CLOCKS->CLK_REF_CTRL = (CLOCKS_CLK_REF_CTRL_SRC_xosc_clksrc << CLOCKS_CLK_REF_CTRL_SRC_Pos);
  CLOCKS->CLK_SYS_CTRL = (CLOCKS_CLK_SYS_CTRL_AUXSRC_clksrc_pll_sys << CLOCKS_CLK_SYS_CTRL_AUXSRC_Pos) |
                         (CLOCKS_CLK_SYS_CTRL_SRC_clksrc_clk_sys_aux << CLOCKS_CLK_SYS_CTRL_SRC_Pos);

  CLOCKS->CLK_PERI_CTRL = CLOCKS_CLK_PERI_CTRL_ENABLE_Msk |
      (CLOCKS_CLK_PERI_CTRL_AUXSRC_clk_sys << CLOCKS_CLK_PERI_CTRL_AUXSRC_Pos);

  CLOCKS->CLK_ADC_CTRL = CLOCKS_CLK_ADC_CTRL_ENABLE_Msk |
      (CLOCKS_CLK_ADC_CTRL_AUXSRC_xosc_clksrc << CLOCKS_CLK_ADC_CTRL_AUXSRC_Pos);

  RESETS_CLR->RESET = RESETS_RESET_io_bank0_Msk |             // enable GPIO
                      RESETS_RESET_pads_bank0_Msk;
  while((RESETS->RESET_DONE & (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk)) !=
        (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk));
}


// stabile Zustaende ohne HW-Reset
void reset_io(void)
{
// reset core1
  PSM_SET->FRCE_OFF = PSM_FRCE_OFF_proc1_Msk;       // core1 passiv
  while(!(PSM->FRCE_OFF & PSM_FRCE_OFF_proc1_Msk)); // abwarten
  PSM_CLR->FRCE_OFF = PSM_FRCE_OFF_proc1_Msk;       // Bit wieder loeschen
  
// reset diverse Peripherie beim programmstart
  RESETS_SET->RESET = RESETS_RESET_io_bank0_Msk | 
                      RESETS_RESET_pwm_Msk | 
                      RESETS_RESET_spi1_Msk | 
                      RESETS_RESET_dma_Msk | 
                      RESETS_RESET_pio0_Msk |
                      RESETS_RESET_pio1_Msk |
                      RESETS_RESET_pads_bank0_Msk;
  SIO->GPIO_OE_CLR = 0xffffffff;  // alle loeschen
  short_delay();  
}

void lade_rom_daten(uint32_t offset)              // gespeicherte Daten laden
{
  int n;
  uint8_t *p = (uint8_t *) (STD_ROM_DATEN + offset);
  IO_BANK0->GPIO25_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // LED-Ausgang
  SIO->GPIO_OE_SET = BIT(LOCAL_LED); 
  SIO->GPIO_OUT_SET = BIT(LOCAL_LED);             // on; zur Kontrolle
   
  for(n = 0; n < MAX_ROM_SIZE; n++) {
    rom_puffer[n] = *p++; 
  }
    SIO->GPIO_OUT_CLR = BIT(LOCAL_LED);           // off
  }  

int main(void)
{
  reset_io();
  init_clock();

  rom2ram();                        // umkopieren und main() im RAM ausführen
  lade_rom_daten(0);                // immer offset = 0 als Startwert

  init_PIO0_RX();                   // mit 500 kB empfangen, Monoflop fuer Reset
  init_PIO1();                      // Ausgabe eines Bytes bei Adressaenderung
  trigger_reset();                  // Reset ausgeben
  rom_simulation(MAX_ROM_SIZE-1);   // endlos 32 k

  while(1) {}
}  
