/*
PIO0_IRQ_0_IRQHandler(void):
Hier werden RX-Daten/Befehle empfangen und im Interrupt verarbeitet.
Daten werden im Intel-HEX-Format oder als Motorola S-Records erwartet.

Zusaetzliche Befehle koennen bei Bedarf die Ablageadressen mit dem Befehl
=Sxxxx maskieren, wobei xxxx die eff. Groesse angibt. Als Maske wird xxxx-1 verwendet.
Ein Offset zur Ablage im ROM-Puffer wird mit =Oxxxx angegeben.

Der Inhalt des ROM-Puffers kann gespeichert werden: =R{index}<CR>
Der Speicherindex ist optional und kann 0 - 7 betragen, wodurch die Ablage von 
max. 8 verschiedenen Versionen angewaehlt werden kann. Default: index = 0.
Die eff. Speicheradresse im FLASH ist: ROM_DATEN_START + MAX_ROM_SIZE * index.

Beim Programmstart werden die Daten ab ROM_DATEN_START in den ROM-Puffer geladen.
Alternativ koennen indiziert andere Daten geladen werden: =L{index}<CR>
Die eff. Laderadresse im FLASH ist: ROM_DATEN_START + MAX_ROM_SIZE * index.

2025-04-21
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#include "rom_simul.h"

#define LF            10
#define CR            13
#define SREC_START    'S'
#define IHEX_START    ':'
#define CMD           '='
#define ROM_SIZE      'S'
#define ROM_OFFSET    'O'
#define ROM_RETTEN    'R'                         // Retten
#define ROM_LADEN     'L'                         // Laden
#define RESERVE_BYTES 50                          // Puffer vor Ueberlauf schuetzen

#define RX_PUF_LEN    200
#define BIN_PUF_LEN   100

extern uint32_t SystemCoreClock;
extern void lade_rom_daten(uint32_t offset);      // indiziertes Laden
extern void speicher_rom_puffer(uint32_t offset); // indiziertes Speichern

void bewerte_srec(void);
void bewerte_ihex(void);

uint8_t rom_puffer[MAX_ROM_SIZE + RESERVE_BYTES];
uint32_t  rom_size = MAX_ROM_SIZE,
          rom_maske = MAX_ROM_SIZE-1,
          rom_offset = 0;

int fehler, count;

uint8_t rx_puffer[RX_PUF_LEN];
uint8_t bin_puffer[BIN_PUF_LEN];
uint32_t rx_puf_index, rx_puf_anzahl;
uint32_t bin_puf_index, bin_puf_anzahl;
uint32_t temp_adr;


// Endlosschleife
void rom_simulation(uint32_t maske)
{
register uint32_t adresse;
  while(1) {
    adresse = SIO->GPIO_IN & maske;
    PIO1->TXF0 = rom_puffer[adresse];
  }
}  
 
// nachfolgende Routinen lesen ihex- oder srec-Dateien

int wandel_nibble(uint8_t c)
{
  if((c >='0') && (c <= '9')) return(c-'0');
  if((c >='A') && (c <= 'F')) return(c-'A'+10);
  if((c >='a') && (c <= 'f')) return(c-'a'+10);
  return(-1);
}

int lese_byte(uint8_t * s, int *i)
{
int ergebnis=0;
int byte;
  byte = wandel_nibble(s[(*i)++]);
  if(byte < 0) return (-1);       
  ergebnis=byte<<4;
  byte = wandel_nibble(s[(*i)++]);
  if(byte < 0) return (-1);
  ergebnis|=byte;
  return(ergebnis);
}

int lese_hex(uint8_t *s)
{
  uint32_t ergebnis = 0, temp = 0;
  while(1) {
    temp = wandel_nibble(*s++);
    if(temp == -1) break;
    ergebnis <<= 4;
    ergebnis += temp;
  }
  return ergebnis;
}  


int wandle_s_zeile(uint8_t *s, uint8_t *ablage)
{
int i=0, mode=0;
int wert,anzahl;
  if(s[i++] != 'S') return(-1);	          // kein S-record 
  mode = s[i++];
  anzahl = lese_byte(s,&i);
  if(anzahl >= 0) *ablage++= anzahl;
  else return(-1);
  while(anzahl--) {
    if((wert=lese_byte(s,&i)) >= 0) *ablage++= wert;
    else return(-1);
  }
  return(mode);
}



int split_s_record(uint8_t *s, uint8_t *ablage, uint8_t *s_mode, uint8_t *len, uint32_t *adr)
{
uint8_t qsum;
uint8_t n, i, anzahl;
uint32_t adr_temp=0l;
  *adr = adr_temp;
  *len = 0;
  *s_mode = wandle_s_zeile(s,ablage);	    // nibble wandeln
  if(*s_mode < '0') return(1);
  anzahl=ablage[0];
  *len=ablage[0]-3;	                      // rohdaten ohne len, adr, quersumme bei s0, s1, s9
  qsum=0;
  for(i=0;i<=anzahl;i++) {		            // quersumme testen
    qsum += ablage[i];
  }
  if(qsum != 0xff) return(qsum);

  i=1;
  switch(*s_mode) {
    case '2':		                          // bei langen Adressen 
    case '8':
	    (*len)--;	                          // wg. laengerer Adresse 
	    adr_temp += (uint8_t)ablage[i++];
    case '0':	                            // keine Nutzdaten, sondern Befehle: ignorieren
    case '1':
    case '9':
	    adr_temp <<= 8;
	    adr_temp += (uint8_t)ablage[i++];
	    adr_temp <<= 8;
	    adr_temp += (uint8_t)ablage[i++];
  }
  n=0;
  while(i<anzahl) {
    ablage[n++] = ablage[i++];	          // nutzdaten ab [0] legen 
  }
  ablage[n]=0;		                        // falls String ohne Abschluss 
  *adr = adr_temp;
  return(0);
}

void bewerte_srec(void)
{
uint8_t s_mode, len, i;
uint32_t adr;
static int status;

  status=split_s_record(rx_puffer, bin_puffer, &s_mode, &len, &adr);
  if(status == 0) {
    adr &= rom_maske;                     // nach Vorgabe maskieren
    adr += rom_offset;                    // und / oder mit Offset ablegen
    adr &= MAX_ROM_SIZE-1;                // auf max. Bereich begrenzen
    for(i=0; i<len; i++)    
      rom_puffer[adr + i] = bin_puffer[i];
  }
  else  fehler++; 
}  


int wandle_ihex_zeile(uint8_t *s, uint8_t *ablage)
{
int i=0;
int wert,anzahl;
if(s[i++] != ':') return(-1);	            // kein ihex-record 
  anzahl = lese_byte(s,&i);
  if(anzahl >= 0) *ablage++= anzahl;
  else return(-1);
  anzahl += 4;                            // gesamte Anzahl 
  while(anzahl--) {
    if((wert=lese_byte(s,&i)) >= 0) *ablage++= wert;
    else return(-1);
  }
  return(0);
}



int split_ihex_record(uint8_t *s, uint8_t *ablage, uint8_t *s_mode, uint8_t *len, uint32_t *adr)
{
uint8_t qsum;
uint8_t i, anzahl;
uint32_t adr_temp=0l;
  *adr = adr_temp;
  *s_mode = wandle_ihex_zeile(s,ablage);	// nibble wandeln
  if(*s_mode != 0) return(1);             // wird nicht verarbeitet
  *len = anzahl = ablage[0];
  anzahl += 4;                            // gesamte Anzahl 
  qsum=0;
  for(i=0; i<=anzahl;i++) {		            // alle Bytes addieren
    qsum += ablage[i];
  }
  if(qsum) return(-1);

  adr_temp += (uint8_t)ablage[1];         // Adresse zusammensetzen
  adr_temp <<= 8;
  adr_temp += (uint8_t)ablage[2];
  *adr = adr_temp;
  return(0);
}

void bewerte_ihex(void)
{
uint8_t s_mode, len, i;
static uint32_t adr;
static int status;

  status=split_ihex_record(rx_puffer, bin_puffer, &s_mode, &len, &adr);
  if(status == 0 && bin_puffer[3] == 0) {
    adr &= rom_maske;                     // nach Vorgabe maskieren
    adr += rom_offset;                    // und / oder mit Offset ablegen
    adr &= MAX_ROM_SIZE-1;                // auf max. Bereich begrenzen
    temp_adr = adr;

    for(i=0; i<len; i++)
      rom_puffer[adr + i] = bin_puffer[i+4];
  }
  else  fehler++;
  
}  

void PIO0_IRQ_0_IRQHandler(void)
{
uint8_t wert, befehl; 
uint32_t temp;
  if(!(PIO0->IRQ0_INTS & PIO0_INTR_SM0_RXNEMPTY_Msk)) return; // nichts zu tun

//RX-wert empfangen
  PIO0->IRQ0_INTF &= ~PIO0_INTR_SM0_RXNEMPTY_Msk;
  wert = (PIO0->RXF0 >> 24);
  switch(wert) {
  case LF:
    break;
  case CR:
    befehl = rx_puffer[0];
    rx_puffer[rx_puf_index] = 0;                  // abschluss
    rx_puf_anzahl = rx_puf_index;
    switch(befehl) {
      case SREC_START:                            // SREC lesen
        bewerte_srec();
        break;
      case IHEX_START:                            // IHEX lesen
        bewerte_ihex();
        break;
      case CMD:                                   // Befehl holen
        befehl = toupper((int)rx_puffer[1]);
        switch(befehl) {  
          case ROM_SIZE:                          // Adresse maskieren
            temp = lese_hex(&rx_puffer[2]);
            if(temp <= MAX_ROM_SIZE &&            // passt ins ROM
               temp >= MIN_ROM_SIZE &&            // mindesten 4 K
              (temp & (MIN_ROM_SIZE-1)) == 0) {   // und passende Adresse
              rom_size = temp;
              rom_maske= rom_size - 1;
            }  
          break;
          case ROM_OFFSET:                        // Adressoffset holen
            temp = lese_hex(&rx_puffer[2]);
            if(temp <= MAX_ROM_SIZE &&            // passt ins ROM
               temp >= MIN_ROM_SIZE &&            // mindestens 4 K
              (temp & (MIN_ROM_SIZE-1)) == 0) {   // und passende Adresse
              rom_offset = lese_hex(&rx_puffer[2]);
            }  
          break;  
          case ROM_RETTEN:                        
            temp = rx_puffer[2] - '0';            // optionalen Speicherindex holen
            if(temp > 7) temp = 0;                // 0 bei ungueltigem Wert
            temp *= MAX_ROM_SIZE;                 // eff. Startoffset
            speicher_rom_puffer(temp);
          break;  
          case ROM_LADEN:                         
            temp = rx_puffer[2] - '0';            // optionalen Ladeindex holen
            if(temp > 7) temp = 0;                // 0 bei ungueltigem Wert
            temp *= MAX_ROM_SIZE;                 // eff. Startoffset
            lade_rom_daten(temp);
          break;  
        }
        break;
    }  
    rx_puf_anzahl = rx_puf_index = 0;
    break;

  default:
    rx_puffer[rx_puf_index++] = wert;             // im Empfangspuffer ablegen
    if(rx_puf_index >= RX_PUF_LEN) rx_puf_index = 0;
  }
}  
  
